class = require "lib.middleclass.middleclass"
require "lib.HardonCollider"

--[[
-- Base class for any drawable object
--]]
GameObject = class("Game Object")

--[[
-- x, y: world coords where to locate the image origin
-- wrap: if true the object wraps around the screen edges
-- v: translational velocity in the direction of 'a' angle
-- a: angle of rotation and heading (in rads)
-- av: angular velocity (rate of rotation)
-- i: image associated with this object
-- hc: hardoncollider instance
]]--
function GameObject:initialize(x, y, wrap, v, a, av, i, hc)
  self.x = x
  self.y = y
  self.wrap = wrap
  self.w = i:getWidth()
  self.h = i:getHeight()
  self.v = v
  self.a = a
  self.av = av
  self.i = i
  self.ox = self.w / 2
  self.oy = self.h / 2
  self.collisionShape = hc.rectangle(x-self.ox, y-self.ox, self.w, self.h)
  self.collisionShape:rotate(self.a)
end

-- Common update housekeeping tasks
function GameObject:update(dt)
  self.a = self.a + self.av

  -- Fix up the angle to only allow values between 0 and 2*pi
  if self.a < 0 then
    self.a = self.a + 2 * math.pi
  elseif self.a > 2 * math.pi then
    self.a = self.a - 2 * math.pi
  end

  -- Fix up coords to keep object in the window
  if self.wrap then
    if self.x < 0 then self.x = WINDOW_WIDTH-1 end
    if self.x >= WINDOW_WIDTH then self.x = 0 end
    if self.y < 0 then self.y = WINDOW_HEIGHT-1 end
    if self.y >= WINDOW_HEIGHT then self.y = 0 end
  end

  self.collisionShape:rotate(self.av)
  self.collisionShape:moveTo(self.x, self.y)
end


local function getEdge (pos, rad, max)
  if pos - rad < 0 then return pos + max end
  if pos + rad > max then return pos - max end
end

function GameObject:drawAt(x, y)
  -- Rotate the image around its center. This also moves the origin to the center of the image.
  love.graphics.draw(self.i, x, y, self.a, 1, 1, self.ox, self.oy)
  if DEBUG then
    love.graphics.rectangle("line", x - self.ox, y - self.ox, self.w, self.h)
    local cr, cg, cb, ca = love.graphics.getColor()
    love.graphics.setColor(255,0,0)
    self.collisionShape:draw("line")
    love.graphics.setColor(cr, cg, cb, ca)
  end
end

function GameObject:draw()
  local x = getEdge(self.x, self.w, WINDOW_WIDTH)
  local y = getEdge(self.y, self.h, WINDOW_HEIGHT)

  self:drawAt(self.x, self.y)           -- draw at real position

  if self.wrap then
    if x then self:drawAt(x, self.y) end  -- draw horizontal wrap
    if y then self:drawAt(self.x, y) end  -- draw vertical wrap
    if x and y then self:drawAt(x, y) end -- draw corner wrap
  end

  if DEBUG then
    local cr, cg, cb, ca = love.graphics.getColor()
    love.graphics.setColor(255,0,0)
    self.collisionShape:draw("line")
    love.graphics.setColor(cr, cg, cb, ca)
  end

end

--[[
-- Return true if point is inside a rectangle.
-- Rectangle coords are the top left corner of the rectangle at rx, ry
-- rw: rectangle width
-- rh: rectangle height
--]]
local function pointIsInsideRect(x, y, rx, ry, rw, rh)
  return x > rx and x < rx + rw and y > ry and y < ry + rh
end

--[[
-- Determine whether point (x, y) is within the square keepout area around
-- the object's location. Object's location is its center.
-- Evaluate with wrapping in mind.
-- Return true if the point is within the keepout. False otherwise.
--]]
function GameObject:pointIsInsideKeepout(x, y, keepout)
  local px = getEdge(self.x, keepout, WINDOW_WIDTH)
  local py = getEdge(self.y, keepout, WINDOW_HEIGHT)

  inside = pointIsInsideRect(x, y, self.x - keepout, self.y - keepout, keepout * 2, keepout * 2)

  if px then
    inside = inside or pointIsInsideRect(x, y, px - keepout, self.y - keepout, keepout * 2, keepout * 2)
  end
  if py then
    inside = inside or pointIsInsideRect(x, y, self.x - keepout, py - keepout, keepout * 2, keepout * 2)
  end
  if px and py then
    inside = inside or pointIsInsideRect(x, y, px - keepout, py - keepout, keepout * 2, keepout * 2)
  end

  return inside
end
