--[[
-- Define a class for high score: HighScore
-- Store and retreive high scores in a file: HighScoreStore
--]]

class = require "lib.middleclass.middleclass"

HighScoreStore = class("HighScoreStore")
HighScore = class("HighScore")


-- helper function for splitting strings
-- Doesn't work on all patterns.
-- see: http://stackoverflow.com/questions/1426954/split-string-in-lua
local function split(s, delimiter)
    result = {};
    for match in (s..delimiter):gmatch("(.-)"..delimiter) do
        table.insert(result, match);
    end
    return result;
end

function HighScore:initialize(name, score)
  self.name = name
  self.score = score
end


function HighScoreStore:initialize(fn, maxNumOfScores)
  self.fn = fn
  self.hscores = {}
  self.maxNumOfScores = maxNumOfScores
end

--[[
-- Open the highscore file, read and return all entries in a
-- sorted (descending) table of HighScore objects.
--]]
function HighScoreStore:get()
  -- clean any existing scores
  for i, v in pairs(self.hscores) do
    self.hscores[i] = nil
  end
  
  if not love.filesystem.exists(self.fn) then
    return {}
  end
  
  --[[
  -- File format:
  --  One score per line in tab-delimited format.
  --  Sorted highest score to lowest.
  --  <name>\t<score>\n
  --]]
  for line in love.filesystem.lines(self.fn) do
    local s = split(line, "\t")
    if #s == 2 and #s[1] > 0 and #s[2] > 0 then
      table.insert(self.hscores, HighScore:new(s[1], tonumber(s[2])))
    end
  end
  
  -- Scores should be sorted in the file, but sort them in case a
  -- kid edited the file by hand and didn't bother to sort the entries
  -- manually.
  table.sort(self.hscores, function(a,b) return a.score > b.score end)
  
  return self.hscores
 end

--[[
-- Insert a high score into the highscore list. Store in the high score file.
-- Note that only self.maxNumOfScores are stored. If the score list is
-- already maxed out and the score being inserted is lower than the
-- lowest score in the table, the score being inserted will not be
-- added to the high score table and will not be stored in the file.
--]]
function HighScoreStore:add(highscore)
  table.insert(self.hscores, highscore)
  table.sort(self.hscores, function(a,b) return a.score > b.score end)
  if #self.hscores > self.maxNumOfScores then
    table.remove(self.hscores, #self.hscores)
  end
  
  local f = love.filesystem.newFile("highscores.txt")
  if not f:open("w") then
    return {}
  end
  
  for i, hs in pairs(self.hscores) do
    f:write(hs.name .. "\t" .. hs.score .. "\n")
  end
  
  f:close()
end
