--[[
-- Game of asteroids.
--
-- F1 pauses the game and brings up a Lua console.
--  DEBUG_PLAYER_IS_GOD = true  makes player invulnerable.
-- CTRL+C exits the debug Lua console.
--]]

REVISION = "v1.0"

require "player"
require "asteroid"
require "ufo"
require "bullet"
require "explosion"
require "hud"
require "menu"
require "gameover"
require "stars"
require "heartbeat_audio"

hc = require "lib.HardonCollider"
suit = require "lib/SUIT"

-- Set to true to enable drawing of image objects in white and collision outlines in red
DEBUG = false
DEBUG_NO_ASTEROIDS = false
DEBUG_PLAYER_IS_GOD = false
DEBUG_START_WITH_GAMEOVER = false

-- TODO: Move these into the explosion class?
MAX_NUM_EXPLOSION_PARTICLES = 20
EXPLOSION_TIME_SEC = 1

TIME_BETWEEN_WAVES_SEC = 4

MAX_NUM_EXTRA_LIVES = 4
POINTS_FOR_EXTRA_LIFE = 1000

RESPAWN_TIMER_SEC = 3

--[[
-- Gamestate: "menu", "game", "pause", "gameover"
--]]
local gamestate = "menu"

--[[
-- Populate game world with asteroids depending on the wave level.
-- Asteroids are located away from the player's current position.
--
-- level : wave level 1 (start) to ?
-- player : player object
--
-- returns a table of Asteroid objects.
--]]
local function startNewWave(level, player)
  local keepout = 200 -- 200 pixel box around the player is a keepout. No asteroids.
  local asteroids = {}
  numSmallAsteroidsAtStartofWave = 0

  -- Number of asteroids per level is level + 2
  for i=1, level do
    local x = math.random(0, WINDOW_WIDTH-1)
    local y = math.random(0, WINDOW_HEIGHT-1)

    -- Ensure that asteroid coords are outisde of the player keepout.
    -- This is a terrible way of doing this. Could just add/subtract some
    -- portion of the keepout and check again.
    while player:pointIsInsideKeepout(x, y, keepout) do
      x = math.random(0, WINDOW_WIDTH-1)
      y = math.random(0, WINDOW_HEIGHT-1)
    end

    local a = math.random() * math.pi * 2
    -- asteroid velocity becomes generally faster with levels
    local v = 10 * level + math.random() * 100
    local av = math.pi / 256 + math.random() * math.pi / 64
    if (math.random() < 0.5) then av = av * -1 end
    local s = math.random(0, 100)
    if (s < 15) then
      size_str = "small"
      numSmallAsteroidsAtStartofWave = numSmallAsteroidsAtStartofWave + 1
    elseif (s >= 15 and s < 55) then
      size_str = "medium"
      numSmallAsteroidsAtStartofWave = numSmallAsteroidsAtStartofWave + 4
    else 
      size_str = "large"
      numSmallAsteroidsAtStartofWave = numSmallAsteroidsAtStartofWave + 16
    end

    asteroid = Asteroid:new(x, y, v, a, av, size_str, hc)
    table.insert(asteroids, asteroid)
    numAsteroids = numAsteroids + 1
  end
  
  numSmallAsteroidsRemaining = numSmallAsteroidsAtStartofWave
  backgroundHeartbeat:play(1.0)
  
  return asteroids
end

local function createNewPlayer(horizVelocity)
  return Player:new(
    WINDOW_WIDTH / 2,                           -- x
    WINDOW_HEIGHT / 2,                          -- y
    horizVelocity,                              -- v (velocity)
    love.graphics.newImage("asset/img/ship_44x44.png"),   -- player image
    love.graphics.newImage("asset/img/thrust_16x16.png"), -- thrust exhaust image
    love.audio.newSource("asset/audio/thrust.wav", "static"), -- thrust sound
    love.audio.newSource("asset/audio/player_shot.wav", "static"), -- shot sound
    asteroidDestructSound, -- player destruction sound
    hc)                                         -- collider
end

local function destroyPlayer()
  player:destroy()

  table.insert(explosions, Explosion:new(
      player.x,
      player.y,
      {249, 100, 100, 255},
      MAX_NUM_EXPLOSION_PARTICLES*20,
      EXPLOSION_TIME_SEC*2))

  scoreTowardsNextLife = 0
  if numExtraLives > 0 then
    numExtraLives = numExtraLives - 1
  end

  respawnTimer = RESPAWN_TIMER_SEC
end

function love.load(arg)
  if arg[#arg] == "-debug" then require("mobdebug").start() end  

  math.randomseed(os.time())

  WINDOW_WIDTH = love.graphics.getWidth()
  WINDOW_HEIGHT = love.graphics.getHeight()

  menu = Menu:new()
  gameover = GameOver:new()

  stars = Stars:new(100, WINDOW_WIDTH, WINDOW_HEIGHT, 3)

  hud = Hud:new()
  
  asteroidDestructSound = love.audio.newSource("asset/audio/asteroid_explosion.wav", "static")
  ufoEntersSound = love.audio.newSource("asset/audio/ufo_enters.wav")


  -- create a player that is moving horizontally across the screen for the menu
  player = createNewPlayer(70)
  bullets = {}
  ufoBullets = {} -- UFO bullets. Bad for player.
  asteroids = {}
  -- since asteroids table is sparsly populated, #asteroids doesn't work. 
  -- keep a running tally of asteroids in numAsteroids
  numAsteroids = 0 
  numSmallAsteroidsAtStartofWave = 0
  numSmallAsteroidsRemaining = 0
  explosions = {}
  ufos = {}

  numExtraLives = 2
  respawnTimer = RESPAWN_TIMER_SEC

  waveNum = 1
  betweenWavesTimer = TIME_BETWEEN_WAVES_SEC

  gamestate = "menu"
  prevGamestate = "init"

  backgroundHeartbeat = HeartbeatAudio:new(love.audio.newSource("asset/audio/heartbeat.wav", static))

  score = 0
  prevScore = 0
  scoreTowardsNextLife = 0
end

local function destroyAsteroid(idx)
  if asteroids[idx].s == "small" then
    numSmallAsteroidsRemaining = numSmallAsteroidsRemaining - 1
    backgroundHeartbeat:setRate(1 + 1 - numSmallAsteroidsRemaining / numSmallAsteroidsAtStartofWave)
    asteroids[idx] = nil
    numAsteroids = numAsteroids - 1
    score = score + 30
    return
  end

  local new_size
  if asteroids[idx].s == "medium" then
    new_size = "small"
    score = score + 20
  else
    new_size = "medium"
    score = score + 10
  end

  local x = asteroids[idx].x
  local y = asteroids[idx].y

  -- remove the bigger asteroid that just got broken up into smaller chunks
  asteroids[idx] = nil
  numAsteroids = numAsteroids - 1

  -- spawn 4 asteroids in place of the destroyed one
  for i = 0, 3 do
    local a = math.random() * math.pi * 2
    local v = 10 * waveNum + math.random() * 100
    local av = math.pi / 256 + math.random() * math.pi / 64
    if (math.random() < 0.5) then av = av * -1 end
    local asteroid = Asteroid:new(x, y, v, a, av, new_size, hc)
    table.insert(asteroids, asteroid)
    numAsteroids = numAsteroids + 1
  end
end

local function destroyUfo(idx)
  ufos[idx] = nil
  score = score + 50
end

local function cleanUpGameObjects()
  numAsteroids = 0
  asteroids = {}
  ufos = {}
  ufoBullets = {}
  bullets = {}
  explosions = {}
  score = 0
  prevScore = 0
  scoreTowardsNextLife = 0
  prevUfoSpawnScore = 0
  waveNum = 1
  betweenWavesTimer = TIME_BETWEEN_WAVES_SEC
end


function love.update(dt)
  if DEBUG_START_WITH_GAMEOVER then
    gamestate = "gameover"
  end

  if gamestate == "game" then
    if prevGamestate == "menu" then
      asteroids = startNewWave(waveNum, player)
      numExtraLives = 2
    end
    backgroundHeartbeat:update(dt)
    
    player:update(dt)

    for ei, explosion in pairs(explosions) do
      if explosion.timer > 0 then
        explosion:update(dt)
      else
        explosions[ei] = nil
      end
    end

    if player.destroyed then
      respawnTimer = respawnTimer - dt
      if respawnTimer <= 0 then
        if numExtraLives > 0 then
          player = createNewPlayer(0)
        else
          gameover:setScore(score)
          gamestate = "gameover"
        end
      end
    end

    for bi, bullet in pairs(bullets) do
      bullet:update(dt)
      if bullet.travelled_too_far then
        bullets[bi] = nil
      end
    end

    for ubi, ufoBullet in pairs(ufoBullets) do
      ufoBullet:update(dt)
      if ufoBullet.travelled_too_far then
        ufoBullets[ubi] = nil
      end
    end

    -- update each asteroid. check for collision with bullets
    for ai, asteroid in pairs(asteroids) do
      asteroid:update(dt)
      for bi, bullet in pairs(bullets) do
        -- check bullet collision with asteroids
        if bullet.collisionShape:collidesWith(asteroid.collisionShape) then
          table.insert(explosions, Explosion:new(
              asteroid.x,
              asteroid.y,
              {249, 244, 102, 255},
              MAX_NUM_EXPLOSION_PARTICLES,
              EXPLOSION_TIME_SEC))
          destroyAsteroid(ai)
          asteroidDestructSound:play()
          bullets[bi] = nil
        end
      end
    end

    -- need a separate loop through asteroids because the previous loop may have
    -- created more asteroids
    for ai, asteroid in pairs(asteroids) do
      if not player.invulnerable and not player.destroyed then
        if player.collisionShape:collidesWith(asteroid.collisionShape) then
          destroyPlayer()
        end
      end
    end

    -- check for UFO bullet collisions with the player
    for ubi, ufoBullet in pairs(ufoBullets) do
      if not player.invulnerable and not player.destroyed then
        if player.collisionShape:collidesWith(ufoBullet.collisionShape) then
          destroyPlayer()
        end
      end
    end

    -- Initialize previous UFO spawn score
    if not prevUfoSpawnScore then
      prevUfoSpawnScore = 0
    end
    -- Every 100 points or so there's a 50% chance to spawn a UFO
    if score - prevUfoSpawnScore >= 100 and #ufos == 0 then
      prevUfoSpawnScore = score
      if math.random() < 0.5 then
        local y = math.ceil(math.random() * (WINDOW_HEIGHT - 1))
        -- v is negative because it's moving from right to left
        local v = -100 - (math.random() * 100)
        -- UFOs become more accurate reaching the max accuracy at level 5
        ufo = Ufo:new(WINDOW_WIDTH-1, y, v, 
          love.graphics.newImage("asset/img/ufo_32x32.png"),
          love.audio.newSource("asset/audio/ufo_shot.wav"),
          hc, true, waveNum / 5)
        table.insert(ufos, ufo)
        ufoEntersSound:play()
      end
    end

    -- Update UFO. Check for collision with bullets and player
    if #ufos > 0 then
      for ui, ufo in pairs(ufos) do
        ufo:update(dt, player.x, player.y, player.vx, player.vy)
        -- if UFO left the window through the left edge, destroy it without increasing score
        if ufo.destroy then
          ufos[ui] = nil
        elseif not player.invulnerable and not player.destroyed then
          if player.collisionShape:collidesWith(ufo.collisionShape) then
            destroyPlayer()
          end
        end
        for bi, bullet in pairs(bullets) do
          -- check bullet collision with UFOs
          if bullet.collisionShape:collidesWith(ufo.collisionShape) then
            table.insert(explosions, Explosion:new(ufo.x, ufo.y, {249, 244, 102, 255}, MAX_NUM_EXPLOSION_PARTICLES, EXPLOSION_TIME_SEC))
            destroyUfo(ui)
            bullets[bi] = nil
            asteroidDestructSound:play()
          end
        end
      end
    end

    -- If no UFOs or asteroids are around, start the timer before spawning the next wave
    if numAsteroids == 0 and #ufos == 0 then
      if betweenWavesTimer <= 0 then
        waveNum = waveNum + 1
        asteroids = startNewWave(waveNum, player)
        betweenWavesTimer = TIME_BETWEEN_WAVES_SEC
      else
        betweenWavesTimer = betweenWavesTimer - dt
      end
    end

    if score > prevScore then
      -- Grant an extra life for every few points
      scoreTowardsNextLife = scoreTowardsNextLife + score - prevScore
      if scoreTowardsNextLife >= POINTS_FOR_EXTRA_LIFE then
        if numExtraLives < MAX_NUM_EXTRA_LIVES then
          numExtraLives = numExtraLives + 1
        end
        scoreTowardsNextLife = scoreTowardsNextLife - POINTS_FOR_EXTRA_LIFE
      end  
    end

    prevScore = score

  elseif gamestate == "menu" then
    if prevGamestate ~= "menu" then
      cleanUpGameObjects()
      backgroundHeartbeat:stop()
    end

    menu:update(dt)

    -- Allow player movement and shooting in the main menu
    player:update(dt)

    for bi, bullet in pairs(bullets) do
      bullet:update(dt)
      if bullet.travelled_too_far then
        bullets[bi] = nil
      end
    end
  elseif gamestate == "gameover" then
    if prevGamestate ~= "gameover" then
      backgroundHeartbeat:stop()
    end
    gameover:update(dt)
  end

  prevGamestate = gamestate
end


function love.draw()
  stars:draw()
  if gamestate == "menu" then
    for bi, bullet in pairs(bullets) do
      bullet:draw()
    end
    player:draw()
    menu:draw()
  elseif gamestate == "game" then
    for bi, bullet in pairs(bullets) do
      bullet:draw()
    end
    for ubi, ufoBullet in pairs(ufoBullets) do
      ufoBullet:draw()
    end
    for ai, asteroid in pairs(asteroids) do
      asteroid:draw()
    end
    for ui, ufo in pairs(ufos) do
      ufo:draw()
    end
    player:draw()

    for ei, explosion in pairs(explosions) do
      explosion:draw()
    end
    hud:draw(score, numExtraLives, powerups)

    if numAsteroids == 0 and #ufos == 0 and betweenWavesTimer > 0 then
      backgroundHeartbeat:stop()
      -- Text for which wave is about to start
      local wave_str = string.format("W A V E   %d", waveNum + 1)
      local default_font = love.graphics.getFont()
      local cr, cg, cb, ca = love.graphics.getColor()
      local font = love.graphics.newFont("asset/font/Atmosphere-Regular.TTF", 28)
      love.graphics.setColor(0, 172, 242)
      love.graphics.setFont(font)
      str_w = font:getWidth(wave_str)
      str_h = font:getHeight()
      love.graphics.print(wave_str, (WINDOW_WIDTH - str_w) / 2, 180)
      -- restore font, color, line settings back to what they were before this call
      love.graphics.setFont(default_font)
      love.graphics.setColor(cr, cg, cb, ca)
    end

  elseif gamestate == "pause" then
  elseif gamestate == "gameover" then
    gameover:draw()
  end

end

function love.keypressed(key)
  if gamestate == "menu" then
    if key == "return" then
      gamestate = "game"
      player = nil
      -- create a stationary player for the game
      player = createNewPlayer(0)
    end
    player:keyPressed(key)
  elseif gamestate == "game" then
    player:keyPressed(key)
  end

  if key == "escape" then
    if gamestate == "menu" then
      love.event.quit()
    else
      gamestate = "menu"
    end
  end

  if key == "f1" then
    debug.debug()
  end

  suit.keypressed(key)
end
-- forward keyboard events
function love.textinput(t)
  suit.textinput(t)
end
