require "game_object"
require "bullet"
require "thrust"

Player = class("Player", GameObject)

Player.static.SHOOT_TIMER_SEC = 0.4
Player.static.INVULNERABILITY_TIMER_SEC = 5.0
Player.static.BULLET_SPEED = 500
Player.static.MAX_BULLET_DISTANCE = 300
Player.static.ACCELERATION = 125

--[[
-- x, y : center coords
-- v : velocity
-- i : player image
-- ti : thrust image (drawn when thrusting)
-- ts : thrust sound
-- ss : shot sound
-- ds : player destruction sound
-- acc : acceleration factor
-- hc: instance of HardonCollider
]]--
function Player:initialize(x, y, v, i, ti, ts, ss, ds, hc)
  GameObject.initialize(self, x, y, true, v, 0, 0, i, hc)
  self.vx = v     -- velocity in x axis
  self.vy = 0     -- velocity in y axis
  self.acc = Player.ACCELERATION
  self.canShoot = true
  self.canShootTimer = Player.SHOOT_TIMER_SEC
  -- Override the default collision shape: a rectangle provided by GameObject
  self.collisionShape = hc.polygon(0, 0, 32, 12, 0, 24)
  self.invulnerabilityTimer = Player.INVULNERABILITY_TIMER_SEC
  self.invulnerable = true
  self.invulnDrawNum = 0
  self.ti = ti
  self.ts = ts
  self.ds = ds
  -- Fix thrusters being too loud. Too lazy to do in Audacity.
  self.ts:setVolume(0.4)
  self.ss = ss
  self.drawThrust = false
  thrust = Thrust:new(x, y, ti, hc)
  self.destroyed = false
end

function Player:update(dt)
  if self.destroyed then
    return
  end

  -- Update ship's angle (the way the ship nose is facing)
  if love.keyboard.isDown("left") or love.keyboard.isDown("a") then
    self.a = self.a - math.pi / 42;
    self.collisionShape:rotate(-math.pi / 42)
  elseif love.keyboard.isDown("right") or love.keyboard.isDown("d") then
    self.a = self.a + math.pi / 42;
    self.collisionShape:rotate(math.pi / 42)
  end

  if love.keyboard.isDown("up")  or love.keyboard.isDown("w") then
    local v0x = self.vx
    local v0y = self.vy
    -- kinematic formula v = v0 + a*t
    self.vx = v0x + self.acc * math.cos(self.a) * dt
    self.vy = v0y + self.acc * math.sin(self.a) * dt

    -- kinematic formula for displacment dx = (v + v0) * t / 2
    self.x = self.x + (self.vx + v0x) * dt / 2
    self.y = self.y + (self.vy + v0y) * dt / 2

    self.drawThrust = true
  else
    -- velocity doesn't change since there's no acceleration and no friction
    -- kinematic formula for displacement dx = v0 * t + 1/2 * a * t^2
    -- is reduced to dx = v0 * t
    self.x = self.x + self.vx * dt
    self.y = self.y + self.vy * dt
    self.drawThrust = false
  end


  if self.drawThrust then
    -- Calculate coords of the center of the rear of the ship
    thrust:update(
      self.x - 11 * math.cos(self.a),
      self.y - 11 * math.sin(self.a),
      self.a)
  end

  if not self.canShoot then
    self.canShootTimer = self.canShootTimer - dt
    if (self.canShootTimer <= 0) then
      self.canShootTimer = Player.SHOOT_TIMER_SEC
      self.canShoot = true
    end
  end

  if self.invulnerable then
    if not DEBUG_PLAYER_IS_GOD then
      self.invulnerabilityTimer = self.invulnerabilityTimer - dt
    end
    if (self.invulnerabilityTimer <= 0) then
      self.invulnerabilityTimer = Player.INVULNERABILITY_TIMER_SEC
      self.invulnerable = false
    end
  end

  GameObject.update(self, dt)
end

function Player:draw()
  if self.destroyed then
    return
  end

  -- If player is invulnerable make the ship blink every fourth draw
  if self.invulnerable then
    self.invulnDrawNum = self.invulnDrawNum + 1
  else
    self.invulnDrawNum = 0
  end

  if self.invulnDrawNum > 7 then
    self.invulnDrawNum = 0
  end

  -- Do not draw the player every 4th draw if player is invulnerable
  if self.invulnerable and self.invulnDrawNum >= 5 and self.invulnDrawNum <= 7 then
    love.graphics.setColor(255, 255, 255, 150)
  end

  GameObject.draw(self)

  if self.invulnerable and self.invulnDrawNum >= 5 and self.invulnDrawNum <= 7 then
    love.graphics.setColor(255, 255, 255, 255)
  end

  if self.drawThrust then
    self.ts:play()
    thrust:draw()
  end

  if DEBUG then
    local cr, cg, cb, ca = love.graphics.getColor()
    love.graphics.setColor(255,0,0)
    self.collisionShape:draw("line")
    love.graphics.setColor(cr, cg, cb, ca)
  end
end

function Player:keyPressed(key)
  if self.destroyed then
    return
  end

  if key == "space" and self.canShoot then
    -- Figure out where the tip of the ship is
    local bullet_x = self.x + (self.w / 2) * math.cos(self.a)
    local bullet_y = self.y + (self.h / 2) * math.sin(self.a)
    bullet = Bullet:new(
      bullet_x,
      bullet_y,
      self.a,
      love.graphics.newImage("asset/img/bullet_14x4.png"),
      hc,
      Player.BULLET_SPEED,
      Player.MAX_BULLET_DISTANCE)
    table.insert(bullets, bullet)
    self.canShoot = false
    self.ss:play()
  end
end

function Player:startInvulnerabilityTimer()
  self.invulnerabilityTimer = Player.INVULNERABILITY_TIMER_SEC
end

function Player:destroy()
  self.destroyed = true
  self.ds:play()
end

