--[[
-- UFOs travel horizontally across the screen.
-- They do not rotate.
-- They only traverse the screen once.
-- They do not wrap around.
-- Some shoot bullets. Others don't.
--]]

require "game_object"
require "bullet"

Ufo = class("Ufo", GameObject)

Ufo.static.SHOOT_TIMER_SEC = 1.5
Ufo.static.BULLET_SPEED = 200
Ufo.static.MAX_BULLET_DISTANCE = 500

--[[
-- x, y : center coords
-- v : velocity
-- i : image
-- ss : shot sound
-- hc: instance of HardonCollider
-- ableToShoot : true = UFO shoots at player, false = UFO doesn't shoot at player
-- accuracy: 0.0 = random shots, 1.0 = predictive player location (i.e. deadly accurate)
--]]
function Ufo:initialize(x, y, v, i, ss, hc, ableToShoot, accuracy)
  GameObject.initialize(self, x, y, false, v, 0, 0, i, hc)
  self.destroy = false
  self.ss = ss
  self.ableToShoot = ableToShoot
  self.canShoot = false
  self.canShootTimer = Ufo.SHOOT_TIMER_SEC
  self.accuracy = accuracy
end

--[[
-- Return solutions for quadratic
--]]
local function quad(a,b,c)
  local sol
  if (math.abs(a) < 1e-6) then
    if (math.abs(b) < 1e-6) then
      if (math.abs(c) < 1e-6) then
        sol = {0, 0}
      end
    else
      sol = {-c/b, -c/b}
    end
  else
    local disc = b*b - 4*a*c
    if (disc >= 0) then
      disc = math.sqrt(disc)
      a = 2*a
      sol = {(-b-disc)/a, (-b+disc)/a}
    end
  end
  return sol
end


--/**
-- * Based on this: http://stackoverflow.com/questions/2248876/2d-game-fire-at-a-moving-target-by-predicting-intersection-of-projectile-and-u
-- *
-- * Return the firing solution for a projectile starting at 'src' with
-- * velocity 'v', to hit a target, 'dst'.
-- *
-- * @param Object src position of shooter
-- * @param Object dst position & velocity of target
-- * @param Number v   speed of projectile
-- * @return Object Coordinate at which to fire (and where intercept occurs)
-- *
-- * E.g.
-- * >>> intercept({x:2, y:4}, {x:5, y:7, vx: 2, vy:1}, 5)
-- * = {x: 8, y: 8.5}
-- */
local function intercept(src, dst, v)
  local tx = dst.x - src.x
  local ty = dst.y - src.y
  local tvx = dst.vx
  local tvy = dst.vy

  -- Get quadratic equation components
  local a = tvx*tvx + tvy*tvy - v*v
  local b = 2 * (tvx * tx + tvy * ty)
  local c = tx*tx + ty*ty

  -- Solve quadratic
  local ts = quad(a, b, c)

  -- Find smallest positive solution
  local sol
  if (ts) then
    local t0 = ts[1]
    local t1 = ts[2]
    local t = math.min(t0, t1);
    if (t < 0) then
      t = math.max(t0, t1);
    end
    
    if (t > 0) then
      sol = { x = dst.x + dst.vx*t, y = dst.y + dst.vy*t }
    end
  end

  return sol;
end

--[[
-- bx, by : bullet coords (where the bullet originates)
-- tx, ty : current target coords 
-- tvx, tvy : current target velocity in x and y
-- accuracy: required accuracy. Number between 0.0 and 1.0 inclusive.
--    0.0 = random and 1.0 = predict target's location as best as possible
--
--]]
function getBulletAngle(bx, by, tx, ty, tvx, tvy, accuracy)
  -- bd_coords  : bullet destination coords
  local bd_coords = intercept(
      {x = bx, y = by},
      {x = tx, y = ty, vx = tvx, vy = tvy},
      Ufo.BULLET_SPEED)
  
  -- If no solution found, double the speed of the UFO bullet
  -- This is hypothetical. The actual speed of the bullet will not
  -- be doubled, but doubling the speed may find a solution that's not
  -- directly pointed at the player but is pretty close to where the
  -- player will be.
  if not bd_coords then
    bd_coords = intercept(
      {x = bx, y = by},
      {x = tx, y = ty, vx = tvx, vy = tvy},
      2 * Ufo.BULLET_SPEED)
    -- If there's still no solution, just shoot at the current target coords
    if not bd_coords then
      bd_coords = { x = tx, y = ty }
    end
  end
  
  -- This is the best we can do without guided bullets
  local kill_angle = math.atan2(bd_coords.y - by, bd_coords.x - bx)
  
  -- Fudge the angle based on accuracy probability
  -- Accuracy of 0.0 means random, 1.0 means the best we can do.
  local rf = math.random()  -- random factor
  
  -- Invert the sign of random angle half the time
  if rf <= 0.5 then
    rf = -rf
  end
  
  -- accuracy = 0.0 : add to kill_angle a number in the range of [-pi .. pi]
  -- accuracy = 0.3 : add to kill_angle a number in the range of [-0.7*pi .. 0.7*pi]
  -- accuracy = 0.5 : add to kill_angle a number in the range of [-0.5*pi .. 0.5*pi]
  -- accuracy = 1.0 : do not add anything to kill_angle
  return kill_angle + rf * (1.0 - accuracy) * math.pi
end

function Ufo:update(dt, player_x, player_y, player_vx, player_vy)
  if self.ableToShoot then
    if self.canShoot then
      local bullet_x = self.x
      local bullet_y = self.y
      local bullet_a = getBulletAngle(bullet_x, bullet_y, player_x, player_y, player_vx, player_vy, self.accuracy)
    
      bullet = Bullet:new(
        bullet_x, bullet_y, bullet_a,
        love.graphics.newImage("asset/img/ufo_bullet_14x4.png"),
        hc, Ufo.BULLET_SPEED, Ufo.MAX_BULLET_DISTANCE)
      table.insert(ufoBullets, bullet)
      self.ss:play()
      self.canShoot = false
    else
      self.canShootTimer = self.canShootTimer - dt
      if (self.canShootTimer <= 0) then
        self.canShootTimer = Ufo.SHOOT_TIMER_SEC
        self.canShoot = true
      end
    end
  end

  self.x = self.x + self.v * dt
  
  -- If UFO is outside of the screen, mark it to be destroyed
  if self.x < 0 or self.x > WINDOW_WIDTH or self.y < 0 or self.y > WINDOW_HEIGHT then
    self.destroy = true
  end
  
  GameObject.update(self, dt)
end